/* $Id$ */
/* load a kernel image, a ramdisk image and jump into the first one.

This file is part of Milo.

Milo is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Milo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Milo; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <linux/types.h>
#include <linux/a.out.h>
#include <asm/bootinfo.h>
#include <milo.h>

/* These are initialized from the box identification process 
   and used by boot() and laucnh() */
unsigned long memory_upper = (unsigned long) NULL;
unsigned long memory_lower = (unsigned long) NULL;
machine_funcp mach_init;	/* pointer to the machine specific initialization */
				/* routine */

/* These are initialized by boot() and launch() use them */
u_long kernel_base = KERNELBASE;
u_long kernel_entry;
u_long kernel_bss_end;

int 
boot (char *kernel_name, char *ramdisk_name)
{
  int rfd = -1;
  struct nlist *nl;
  u_long mem_size;
  u_long kernel_size;
  u_long rd_size;
  char *rdptr;
  u_long temp_ulong;
  char *kptr;

  /*
   * Verify that we got memory info in the tags
   */
  if (memory_upper == (unsigned long) NULL
      || memory_lower == (unsigned long) NULL)
    {
      fprintf (stderr,
	       "Something went wrong with the machine identification (memory), aborting\n\r"
	       "lower: %08x upper: %08x\r\n",
	       (unsigned int) memory_lower,
	       (unsigned int) memory_upper);
      exit (EXIT_FAILURE);
    }

  /*
   * Verify that there is enough RAM
   */
  mem_size = memory_upper - memory_lower;
  if (mem_size < 0x800000)
    {
      fprintf (stderr,
	       "Insufficient Memory to load Linux/MIPS, aborting\n\r");
      exit (EXIT_FAILURE);
    }

  if (option_verbose)
    {
      printf ("%s: kernel file is `%s'\r\n",
	      (behaviour == BEHAVE_PANDORA ? NAMEOF_PANDORA : NAMEOF_MILO),
	      kernel_name);
      if (ramdisk_name)
	printf ("%s: ramdisk file is `%s'\r\n",
		(behaviour == BEHAVE_PANDORA ? NAMEOF_PANDORA : NAMEOF_MILO),
		ramdisk_name);
    }

  /*
   * Open kernel and gather some data from the executable
   */
  if (kld->ld_open_kernel (kernel_name) < 0)
    {
      fprintf (stderr, "Error opening kernel file %s.\n\r", kernel_name);
      exit (EXIT_FAILURE);
    }
  kernel_base = kld->ld_get_kbase ();
  kernel_size = kld->ld_get_ksize ();

  temp_ulong = rd_size = 0;	/* default: no ramdisk */
  if (ramdisk_name)
    {
      if ((rfd = open (ramdisk_name, O_RDONLY, 0)) == -1)
	{
	  fprintf (stderr,
		   "Unable to open ramdisk file %s\n\r", ramdisk_name);
	  exit (EXIT_FAILURE);
	}
      else
	{
	  /*
	   * record ramdisk size
	   */
	  rd_size = lseek (rfd, 0, L_XTND);
	  if (rd_size & ((1 << 10) - 1))
	    {
	      /*
	       * Most probably the file is no image at all or has been
	       * corrupted, so print a warning message.
	       */
	      printf ("Warning: Ramdisk size is not multiple of 1024 bytes.\r\n");
	    }
	  temp_ulong = rd_size >> 10;
	}
    }
  /* had ramdisk tags */
  bi_TagAdd (tag_ramdisk_size, ULONGSIZE, (void *) &temp_ulong);

  /* Ok now put the *LAST* tag */
  bi_TagAdd (tag_dummy, 0, NULL);

  /*
   * Find kernel entry point
   */
  if (!(nl = kld->ld_get_nlist (kld->ld_isymbol ("kernel_entry"))))
    {
      perror ("get_nlist: kernel entry point");
      exit (EXIT_FAILURE);
    }
  else
    {
      kernel_entry = nl->n_value;
      free (nl);
    }

  /*
   * End of bss segment - ramdisk will be placed there
   */
  if (!(nl = kld->ld_get_nlist (kld->ld_isymbol ("_end"))))
    {
      perror ("get_nlist: kernel end");
      exit (EXIT_FAILURE);
    }
  else
    {
      kernel_bss_end = nl->n_value;
      free (nl);
    }

  /*
   * allocate buffers for kernel and ramdisk
   */
  if (!(kptr = (char *) malloc (kernel_size)))
    {
      fprintf (stderr, "Unable to allocate %lu bytes of memory\n\r",
	       kernel_size);
      exit (EXIT_FAILURE);
    }
  memset (kptr, 0, kernel_size);

  if (rd_size)
    {
      if (!(rdptr = (char *) malloc (rd_size)))
	{
	  fprintf (stderr, "Unable to allocate %lu bytes of memory\n\r",
		   rd_size);
	  exit (EXIT_FAILURE);
	}
      memset (rdptr, 0, rd_size);
    }

  /*
   * Should check whether kernel & RAMdisk moving doesn't overwrite
   * the bootstraper during startup.
   */
  if (option_verbose)
    {
      /*
       * The following text should be printed by the loader module
       */
      printf ("\r\n");
      kld->ld_print_stats ();
      printf ("Kernel entry at 0x%08lx\n\r", kernel_entry);

      if (rd_size)
	{
	  printf ("Ramdisk img. at 0x%08lx, size %lu Kbytes\n\r",
		  kernel_bss_end, rd_size >> 10);
	}
    }

  /*
   * Now do the real loading
   */
  if (kld->ld_load_kernel (kptr) < 0)
    {
      fprintf (stderr, "Failed to load kernel executable\r\n");
      free (kptr);
      exit (EXIT_FAILURE);
    }
  kld->ld_close_kernel ();

  /*
   * Load the ramdisk image
   */
  if (rfd != -1 && rd_size != 0)
    {
      if (lseek (rfd, 0, L_SET) == -1)
	{
	  fprintf (stderr, "Failed to seek to beginning of ramdisk file\n\r");
	  free (kptr);
	  exit (EXIT_FAILURE);
	}
      if (debuglevel >= 2)
	printf ("reading %lu bytes of ramdisk to %08lx\r\n", rd_size,
		(unsigned long) rdptr);
      if (read (rfd, rdptr, rd_size) != rd_size)
	{
	  fprintf (stderr, "Failed to read ramdisk file\n\r");
	  free (rdptr);
	  exit (EXIT_FAILURE);
	}
      close (rfd);
    }

  if (option_interactive && behaviour != BEHAVE_PANDORA)
    {
      printf ("\r\n");
      printf ("Hit [Enter] to continue the Linux boot...");
      fflush (stdout);
      getchar ();
      printf ("\r\n");
    }
  if (behaviour != BEHAVE_PANDORA)
    {
      printf ("Launching Kernel...\r\n");
    }

  /*
   * Run machine initialization as late as possible
   */
  if (mach_init != (int (*)(void)) NULL)	/* need some specific intialization stuff ? */
    (*mach_init) ();


  /*
   * copy the screen info structures to the kernel image,
   * then execute the copy-and-go code
   */
  launch (kptr, rdptr, kernel_size, rd_size);

  /*
   * famous last words: not readed
   */
  printf ("Kernel returned -- exiting...\r\n");
  sleep (10);

  return 0;
}
